/* $Id: ax25iface.cpp 321 2009-10-19 14:25:40Z oh2gve $
 *
 * Copyright 2009 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Compile contents only if AX.25 support is enabled.
#ifdef HAVE_LIBAX25



// Own stuff.
#include "cfg.h"
#include "aprspacket.h"
#include "ax25iface.h"

// non-std stuff.
#include <netax25/ax25.h>
#include <netax25/axconfig.h>
#include <netax25/axlib.h>
#include <sys/socket.h>
#include <netinet/if_ether.h>
#include <sys/types.h>
#include <errno.h>
#include <netinet/in.h>

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/string.h>




Ax25Iface::Ax25Iface(wxString const& port, Cfg::ConnectionState const& state, wxString const& owncall):
Connection(state),
rxSocket_(-1), txSocket_(-1), initialized_(true), port_(port), ifaceName_(), owncall_(owncall)
{
   // First we initialize libax25.
   if ( ax25_config_load_ports() == 0 )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): no AX.25 ports available."));
      initialized_ = false;
   }
   
   // Then we get device name.
   if ( initialized_ )
   {
      char* ifname = ax25_config_get_dev(port_.char_str());
      if ( ifname == NULL )
      {
         wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to get AX.25 network interface name."));
         initialized_ = false;
      }
      else
      {
         ifaceName_ = wxString::FromAscii(ifname);
         wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): successfully detected iface name ") + ifaceName_ + wxT(" for the port."));
      }
   }
   
   // Then we open rx socket.
   if ( initialized_ && inputEnabled() )
   {
      // Don't know if this is the best way, but other dudes seem to do it this way also...
      int proto = PF_AX25;
      if ( (rxSocket_ = socket(PF_INET, SOCK_PACKET, htons(proto))) < 0 )
      {
         wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to open rx socket."));
         initialized_ = false;
      }
   }
   
   // Finally we open tx socket.
   if ( initialized_ && outputEnabled() )
   {
      if ( (txSocket_ = socket(PF_AX25, SOCK_DGRAM, 0)) < 0 )
      {
         wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to open tx socket."));
         initialized_ = false;
      }
   }
}


Ax25Iface::~Ax25Iface()
{
   close(rxSocket_);
   close(txSocket_);
}


void* Ax25Iface::Entry()
{
   // Tell main that we're alive.
   threadStarted();
      
   // Check how we did at constructor and try to connect if we're ok.
   if ( initialized_ && openConnection() )
   {
      gotConnected();
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): connection opened."));
   }
   else
   {
      initialized_ = false;
   }
   startupDone();
   
   // If we're going to TX, prepare for it.
   full_sockaddr_ax25 txDstAddr;
   int txDstAddrLen = 0;
   if ( outputEnabled() && (txDstAddrLen = ax25_aton(OWN_TARGET, &txDstAddr)) < 0 )
   { 
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed convert ") + wxT(OWN_TARGET) + wxT(" into network format."));
      return false;
   }
   
   // Main loop of this thread.
   while ( initialized_ && !TestDestroy() )
   {
      // Check if we're supposed to do input.
      if ( inputEnabled() )
      {
         // Prepare to get data from socket.
         sockaddr ifaceAddress;
         size_t addrSize = sizeof(sockaddr);
         char buffer[8192];
         
         // Check for data.
         int bytesRead = recvfrom(rxSocket_, buffer, WXSIZEOF(buffer), MSG_DONTWAIT, &ifaceAddress, &addrSize);
         if ( bytesRead < 0 )
         {
            if ( errno != EAGAIN )
            {
               wxString msg = wxT(" ax25iface(");
               msg += port_;
               msg += wxT("): failed to read from socket, error code ");
               msg << errno;
               msg += wxT(".");
               wxLogMessage(msg);
               gotDisconnected();
               break;
            }
         }
         else if ( bytesRead > 0 )
         {
            // Check where the bytes are from, pick only our iface.
            if ( strcmp(ifaceAddress.sa_data, ifaceName_.ToAscii()) == 0 )
            {
               // Try to parse the bytes into packet.
               AprsPacket* packet = AprsPacket::fromAX25Frame(string(buffer, bytesRead));
               // If successfull, send to main.
               if ( packet != NULL )
               {
                  inputPush(packet);
               }
            }
         }
      }
      
      // Then let's do tx, if needed.
      if ( outputEnabled() )
      {
         // Prepare to send data to socket.
         
         
         // Wait for 1 ms to get the data.
         AprsPacket const* packet = outputPopWait(1);
         if ( packet != NULL )
         {
            string output = packet->getAX25Frame();
            if ( sendto(txSocket_, output.c_str(), output.length(), 0, (sockaddr*)&txDstAddr, txDstAddrLen) < 0 )
            {
               wxString msg = wxT(" ax25iface(");
               msg += port_;
               msg += wxT("): failed to send to socket, error code ");
               msg << errno;
               msg += wxT(".");
               wxLogMessage(msg);
               gotDisconnected();
               delete packet;
               break;
            }
         }
         delete packet;
      }
      else
      {
         // No busylooping plox.
         wxMilliSleep(1);
      }

   } // while (!TestDestroy())
   
   // Tell user that we're dying.
   wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): shutting down."));
   
   // Change our status to disconnected.
   gotDisconnected();

   // Wait for main thread to stop this thread.
   while (true)
   {
      if ( TestDestroy() )
      {
         break;
      }
      
      wxMilliSleep(1);
   }

   return NULL;
}


bool Ax25Iface::openConnection()
{
   // Resolve port callsign from the port name.
   char* portcall = ax25_config_get_addr(port_.char_str());
   if ( portcall == NULL )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to resolve interface callsign."));
      return false;
   }
   
   // Add digipeaters.
   //string destAndDigis (portcall);
   //destAndDigis += " VIA WIDE2-2";
   
   // Create socket address. It is needed to bind txSocket_ to the interface.
   full_sockaddr_ax25 txSrcAddr;
   int txSrcAddrLen = 0;

   // Convert port call into network format and save it as TX source address.
   if ( (txSrcAddrLen = ax25_aton(portcall, &txSrcAddr)) < 0 )
   { 
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed convert ") + wxString::FromAscii(portcall) + wxT(" into network format."));
      return false;
   }
/*   
   // Save own call into blob as source.
   if ( ax25_aton_entry(owncall_.ToAscii(), ifaceAddress.fsa_ax25.sax25_call.ax25_call) < 0 )
   { 
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed convert ") + owncall_ + wxT(" into network format."));
      return false;
   }
*/   
   // Bind socket to the interface.
   if ( bind(txSocket_, (sockaddr*)&txSrcAddr, txSrcAddrLen) < 0 )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to bind to ") + wxString::FromAscii(portcall) + wxT("."));
      return false;
   }
/*   
   // Disable RELAY digipeating.
   int sockopt = 0;
   if ( setsockopt(socket_, SOL_AX25, AX25_IAMDIGI, &sockopt, sizeof(int)) )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): tried to disable RELAY digipeating, but setsockopt() failed."));
      return false;
   }
   
   // Create address blob, which is used to connect to UNPROTO.
   full_sockaddr_ax25 unprotoAddress;
   unprotoAddress.fsa_ax25.sax25_family = AF_AX25;
   unprotoAddress.fsa_ax25.sax25_ndigis = 0;
   string ax25Target (OWN_TARGET);
   ax25Target += "-0";
   if ( ax25_aton_entry(ax25Target.c_str(), unprotoAddress.fsa_ax25.sax25_call.ax25_call) < 0 )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed convert UNPROTO into network format."));
      return false;
   }
   
   // Connect to UNPROTO.
   if ( connect(socket_, (struct sockaddr *)&unprotoAddress, sizeof(full_sockaddr_ax25)) != 0 )
   {
      wxLogMessage(wxT(" ax25iface(") + port_ + wxT("): failed to connect to UNPROTO."));
      return false;
   }
*/   
   return true;
}



/*
bool Ax25Iface::socketIsConnected()
{
   // Test if the socket exists.
   if (socket_ == 0)
   {
      wxLogMessage(wxT(" uplink: socket has disappeared. (wtf?)"));
      return false;
   }

   // Test if the socket is connected.
   if (!socket_->IsConnected())
   {
      logError();
      return false;
   }

   // Check for I/O errors.
   if (socket_->Error())
   {
      logError();
      return false;
   }
   
   // If enabled, check that we have received data recently.
   if ( noTrafficTimeout_ > 0 )
   {
      if ( time(0) - lastTrafficTime_ > noTrafficTimeout_ )
      {
         wxString msg = wxT(" uplink: No traffic received from ");
         msg += currentServer_;
         msg += wxT(" during last ");
         msg << noTrafficTimeout_/60;
         msg += wxT(" minutes, seems that connection has died.");
         
         wxLogMessage(msg);
         return false;
      }
   }
         
   return true;
}


bool Ax25Iface::reconnect()
{
   // If we're having an i/o-error or some crazy blocking error, it might be that the connection
   // has just timed out. So let's try to reconnect.

   if ( socket_->LastError() != wxSOCKET_IOERR &&
        socket_->LastError() != wxSOCKET_WOULDBLOCK )
   {
      return false;
   }
      
   // Keep trying until connected or shutdown request comes.
   while ( !TestDestroy() )
   {
      // If we have tried to connect recently, wait a little before trying again.
      if ( lastTryTime_ + retryInterval_ >= time(0) )
      {
         wxString msg = wxT(" uplink: waiting ");
         msg << (lastTryTime_ + retryInterval_ - time(0));
         msg += wxT(" seconds before trying again.");
         wxLogMessage(msg);
      }
      while ( lastTryTime_ + retryInterval_ >= time(0) )
      {
         if ( TestDestroy() )
         {
            return false;
         }
         wxSleep(1);
      }
      
      // Try to connect to all servers.
      for ( unsigned int i = 0; i < configs_.size(); i++ )
      {
         // Save server number.
         currentServerNo_ = i;
         
         // Set next server to connect to.
         addr_.Hostname(configs_.at(currentServerNo_).host);
         addr_.Service(configs_.at(currentServerNo_).port);

         // Save the chosen server into the string.
         currentServer_ = configs_.at(currentServerNo_).host;
         wxString port = wxT("");
         port << configs_.at(currentServerNo_).port;
         currentServer_ += wxT(":");
         currentServer_ += port;
         
         // Flush outgoing packets to avoid delayed gating.
         const AprsPacket* output = outputPopWait(1);
         while ( output != 0 )
         {
            delete output;
            output = outputPopWait(1);
         }
            
         // Try to connect.
         if ( connect() )
         {
            // We're now successfully connected, which means that we can to try to
            // reconnect sooner than after retryInterval (if required).
            lastTryTime_ = 0;
            
            // Last info from the server has just been received, so we can reset that
            // timer too.
            lastTrafficTime_ = time(0);
         
            return true;
         }
      }
      
      // No luck this time.
      lastTryTime_ = time(0);
      
      // Check that our error has not changed to an unhandleable one.
      if ( socket_->LastError() != wxSOCKET_IOERR &&
           socket_->LastError() != wxrxSocket_WOULDBLOCK )
      {
         return false;
      }
   }
   
   return false;
}
   
   
void Ax25Iface::logError() const
{
   wxSocketError lastError = rxSocket_->LastError();
   wxString error;

   switch (lastError) {
     case wxrxSocket_NOERROR:
      error = wxT("No error happened.");
      break;
     case wxrxSocket_INVOP:
      error = wxT("Invalid operation.");
      break;
     case wxrxSocket_IOERR:
      error = wxT("Input/Output error.");
      break;
     case wxrxSocket_INVADDR:
      error = wxT("Invalid address passed to wxSocket.");
      break;
     case wxrxSocket_INVSOCK:
      error = wxT("Invalid socket (uninitialized).");
      break;
     case wxrxSocket_NOHOST:
      error = wxT("No corresponding host.");
      break;
     case wxrxSocket_INVPORT:
      error = wxT("Invalid port.");
      break;
     case wxrxSocket_WOULDBLOCK:
      error = wxT("The socket is non-blocking and the operation would block.");
      break;
     case wxrxSocket_TIMEDOUT:
      error = wxT("The timeout for this operation expired.");
      break;
     case wxrxSocket_MEMERR:
      error = wxT("Memory exhausted.");
      break;

     default:
      error = wxT("Error 12765. WTF???");
   }

   wxString msg = wxT(" uplink: connection to ");
   msg += currentServer_;
   msg += wxT(" failed: ");
   msg += error;
   
   wxLogMessage(msg);
}
*/

#endif // HAVE_LIBAX25
