/* $Id: cfg.cpp 323 2009-10-19 18:12:40Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen 
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// own stuff.
#include "cfg.h"
#include "filter.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <iostream>
using std::cerr;
using std::endl;
#include <sstream>
using std::istringstream;
#include <cmath>

// non-std stuff.
#include <fap.h>

// wx stuff.
#include <wx/fileconf.h>
#include <wx/wfstream.h>
#include <wx/log.h>
#include <wx/tokenzr.h>
#include <wx/regex.h>


namespace Cfg
{
   /// Functions to be used only inside this module.
   namespace {
      
      /// Converts connection type string into ConnectionType.
      bool parseType(const wxString& input, ConnectionType& type)
      {
         if ( input == wxT("ax25iface") )
         {
            type = AX25IFACE;
            return true;
         }
         else if ( input == wxT("serialtnc") )
         {
            type = SERIALTNC;
            return true;
         }
         else if ( input == wxT("serialkisstnc") )
         {
            type = SERIALKISSTNC;
            return true;
         }
         
         wxLogMessage(wxT("Error: unknown connection type: ") + input);
         return false;
      }
      
      
      /// Converts connection state string into ConnectionState.
      bool parseState(const wxString& input, ConnectionState& state)
      {
         if ( input == wxT("rxonly") )
         {
            state = READONLY;
            return true;
         }
         else if ( input == wxT("txonly") )
         {
            state = WRITEONLY;
            return true;
         }
         else if ( input == wxT("rxtx") )
         {
            state = READWRITE;
            return true;
         }
         
         wxLogMessage(wxT("Error: unknown state: ") + input); 
         return false;
      }


      /// Creates filter element objects from given filter string.
      bool parseFilter(wxString& str, vector<Filter>& results)
      {
         // Used to separate the filter elements from given filter string.
         wxStringTokenizer splitter (str, wxT("|"));
         
         // While there is element strings available.
         while ( splitter.HasMoreTokens() )
         {
            // Get the element string.
            wxString element = splitter.GetNextToken();

            // Filter element to be created.
            Filter newFilter;
                        
            // Separate element parts from each other.
            istringstream partflow ( string(element.ToAscii()) );
            
            // Tmp var for a part.
            string partStd = "";
            while ( partflow >> partStd )
            { 
               // Convert tmp back to wx.
               wxString part = wxString::FromAscii(partStd.c_str());
               
               // Create regexps for part type detection.
               wxRegEx areaReg;
               if ( !areaReg.Compile(wxT("a/(-?[0-9]+.?[0-9]*)/(-?[0-9]+.?[0-9]*)/(-?[0-9]+.?[0-9]*)/(-?[0-9]+.?[0-9]*)")) )
               {
                  wxLogMessage(wxT("Sorry, internal error at ") + wxString::Format(wxT("%s:%d"), __FUNCTION__, __LINE__));
                  return false;
               }
               wxRegEx rangeReg;
               if ( !rangeReg.Compile(wxT("r/(-?[0-9]+.?[0-9]*)/(-?[0-9]+.?[0-9]*)/([0-9]+.?[0-9]*)")) )
               {
                  wxLogMessage(wxT("Sorry, internal error at ") + wxString::Format(wxT("%s:%d"), __FUNCTION__, __LINE__));
                  return false;
               }
               wxRegEx buddyReg;
               if ( !buddyReg.Compile(wxT("b(/[A-Z0-9\\-\\*\\?]+)+")) )
               {
                  wxLogMessage(wxT("Sorry, internal error at ") + wxString::Format(wxT("%s:%d"), __FUNCTION__, __LINE__));
                  return false;
               }
               
               // Check if it is area part.
               if ( areaReg.Matches(part) )
               {
                  // Get the values.
                  wxString latN = areaReg.GetMatch(part, 1);
                  wxString lonW = areaReg.GetMatch(part, 2);
                  wxString latS = areaReg.GetMatch(part, 3);
                  wxString lonE = areaReg.GetMatch(part, 4);
                  
                  // Tmp for area to be created out of this part.
                  Filter::Area newArea;

                  // Convert values to numerics.
                  if ( !latN.ToDouble(&newArea.latN) || !lonW.ToDouble(&newArea.lonW) ||
                       !latS.ToDouble(&newArea.latS) || !lonE.ToDouble(&newArea.lonE) )
                  {
                     wxLogMessage(wxT("Error: invalid area specification \"") + part + wxT("\"."));
                     return false;
                  }
                  
                  // Check that area has no badness.
                  if ( newArea.latN * newArea.latS < 0 ||
                       newArea.lonE * newArea.lonW < 0 ||
                       newArea.latN - newArea.latS <= 0 ||
                       newArea.lonE - newArea.lonW <= 0 )
                  {
                     wxLogMessage(wxT("Sorry, we do not support areas like ") + wxString::Format(wxT("lat %f..%f lon %f..%f."), newArea.latN, newArea.latS, newArea.lonE, newArea.lonW));
                     return false;
                  }
                  
                  // Insert the fresly found area into current filter.
                  newFilter.addArea(newArea);
               }
               // Check if it is range part.
               else if ( rangeReg.Matches(part) )
               {
                  wxString lat = rangeReg.GetMatch(part, 1);
                  wxString lon = rangeReg.GetMatch(part, 2);
                  wxString dist = rangeReg.GetMatch(part, 3);
                  
                  Filter::Range newRange;
                  
                  if ( !lat.ToDouble(&newRange.lat) || !lon.ToDouble(&newRange.lon) ||
                       !dist.ToDouble(&newRange.dist) )
                  {
                     wxLogMessage(wxT("Error: invalid range specification \"") + part + wxT("\"."));
                     return false;
                  }
                  
                  newFilter.addRange(newRange);
               }
               // Check if this is buddy part.
               else if ( buddyReg.Matches(part) )
               {
                  // Check for previous buddy lists.
                  if ( newFilter.hasBuddyPatterns() )
                  {
                     wxLogMessage(wxT("Error: multiple b/:s specified."));
                     return false;
                  }
                  
                  // Create tokenizer to find patterns from the part.
                  wxStringTokenizer buddySplitter (part, wxT("/"));
                  
                  // Check and remove first token (the b for buddy).
                  if ( !buddySplitter.HasMoreTokens() || buddySplitter.GetNextToken() != wxT("b") )
                  {
                     wxLogMessage(wxT("Error: invalid callsign filter \"") + part + wxT("\"."));
                     return false;
                  }
                  
                  // Get patterns.
                  while ( buddySplitter.HasMoreTokens() )
                  {
                     wxString pattern = buddySplitter.GetNextToken();
                     
                     // Convert simple pattern into regexp syntax.
                     pattern.Prepend(wxT("^"));
                     pattern.Replace(wxT("*"), wxT("[A-Z0-9\\-]*"));
                     pattern.Replace(wxT("?"), wxT("[A-Z0-9\\-]?"));
                     pattern.Append(wxT("$"));
                     
                     // Check that it is valid regexp.
                     wxRegEx testReg (pattern);
                     if ( !testReg.IsValid() )
                     {
                        wxLogMessage(wxT("Error: invalid callsign pattern \"") + pattern + wxT("\"."));
                        return false;
                     }
                     
                     // Save the pattern into current filter.
                     newFilter.addBuddyPattern(pattern);
                  }
                  
                  // Check that we got at least one pattern.
                  if ( !newFilter.hasBuddyPatterns() )
                  {
                     wxLogMessage(wxT("Error: invalid callsign filter \"") + part + wxT("\": no patterns found."));
                     return false;
                  }
                  
               }
               // Unknown part.
               else
               {
                  wxLogMessage(wxT("Error: unknown filter specification \"") + part + wxT("\"."));
                  return false;
               }
            } // while (parts available)

            // Check that we got at least one part.
            if ( newFilter.getPartCount() == 0 )
            {
               wxLogMessage(wxT("Error: invalid filter element \"") + element + wxT("\": no parts found."));
               return false;
            }
            
            // Save the freshly created element.
            results.push_back(newFilter);
            
         } // while (elements available)
         
         // Check that we got at least one filter.
         if ( results.empty() )
         {
            wxLogMessage(wxT(" Debug: empty filter string detected."));
         }
         
         return true;
      }
      
      /// Prints generic error message about invalid config value.
      void printErrorInvalid(wxString const& keyword, wxString const& value, unsigned int const connNo = 12765)
      {
         wxString errmsg = wxT("Error: invalid ") + keyword + wxT(" \"") + value + wxT("\"");
         if ( connNo != 12765 )
         {
            errmsg += wxT(" for connection ");
            errmsg << connNo;
         }
         errmsg += wxT(".");
         wxLogMessage(errmsg);
      }
      
      /// Prints generinc error message about missing config value.
      void printErrorNotFound(wxString const& keyword, unsigned int const connNo = 12765)
      {
         wxString errmsg = wxT("Error: no ") + keyword + wxT(" in config file");
         if ( connNo != 12765 )
         {
            errmsg += wxT(" for connection ");
            errmsg << connNo;
         }
         errmsg += wxT(".");
         wxLogMessage(errmsg);
      }
      
      /// Reads uplink related configuration from config object.
      bool readServer(wxFileConfig const& wxConfigObject, Configuration& configuration)
      {
         // Read APRS-IS servers and ports.
         wxString serporlist;
         if ( wxConfigObject.Read(wxT("/uplink/server"), &serporlist) )
         {
            // Split the servers.
            wxStringTokenizer splitter ( serporlist, wxT(" ") );
            while ( splitter.HasMoreTokens() )
            {
               // Get next server.
               wxString serpor = splitter.GetNextToken();
               
               // Separate hostname and port.
               unsigned int colonPos = serpor.Find(':');
               if ( colonPos > 0 && colonPos < serpor.Length()-1 )
               {
                  wxIPV4address tmp;
                  tmp.Hostname(serpor.Mid(0, colonPos));
                  
                  // Convert port to us int.
                  wxString portStr = serpor.Mid(colonPos+1);
                  unsigned long port = 0;
                  if ( !portStr.ToULong(&port) )
                  {
                     wxLogMessage(wxT("Error: invalid port number \"") + portStr + wxT("\" for server ") + tmp.Hostname() + wxT("."));
                     return false;
                  }
                  tmp.Service(port);
                  
                  // Save the server.
                  configuration.uplink.servers.push_back(tmp);
                  configuration.uplink.serverNames.push_back(serpor);
               }
               else               
               {
                  wxLogMessage( wxT("Error: invalid server string \"") + serpor + wxT("\".") );
                  return false;
               }
            }
            
            // Make sure we got at least one server.
            if ( configuration.uplink.servers.empty() )
            {
               wxLogMessage(wxT("Error: empty server list."));
               return false;
            }
            
         }
         else
         {
            printErrorNotFound(wxT("uplink configuration"));
            return false;
         }
         
         
         // Read server state.
         wxString state;
         if ( !(wxConfigObject.Read(wxT("/uplink/state"), &state)) )
         {
            printErrorNotFound(wxT("server connection state"));
            return false;
         }
         else
         {
            if ( !parseState(state, configuration.uplink.serverState) )
            {
               return false;
            }
         }
         // Read no traffic timeout.
         if ( !(wxConfigObject.Read(wxT("/uplink/notraffictimeout"), &configuration.uplink.noTrafficTimeout) ) )
         {
            configuration.uplink.noTrafficTimeout = 0;
         }
         // Change interval into seconds.
         configuration.uplink.noTrafficTimeout *= 60;
         
         // Read retry interval.
         if ( !(wxConfigObject.Read(wxT("/uplink/retryinterval"), &configuration.uplink.retryInterval) ) ||
              configuration.uplink.retryInterval <= 1 )
         {
            wxLogMessage(wxT("Error: uplink retryinterval invalid or not found."));
            return false;
         }
         // Change interval into seconds.
         configuration.uplink.retryInterval *= 60;
         
         // Read filter.
         wxConfigObject.Read(wxT("/uplink/filter"), &configuration.uplink.filter);
         
         // No fail as we're here.
         return true;
      }
      
      /// Reads each connection section from config object.
      bool readConnections(wxFileConfig const& wxConfigObject, Configuration& configuration)
      {
         // Connection counter.
         unsigned int connectionNo = 1;
         // Config file section search pattern.
         wxString pattern = wxT("/connection");
         pattern << connectionNo;
         while ( wxConfigObject.Exists(pattern) )
         {
            // Close config group part from the pattern.
            pattern += wxT("/");
            
            // Temporary config item for connection config to be read. It's never
            // initialized completely, so it would be wise to be carefull later...
            ConnectionCfg newConnCfg;
            
            // Temporary place to store value read from config file.
            wxString cfgString = wxT("");
            long cfgNumber = 0;
            
            // Look for type.
            if ( wxConfigObject.Read(pattern + wxT("type"), &cfgString) )
            {
               // Check that connection type is valid.
               if ( !parseType(cfgString, newConnCfg.type) )
               {
                  printErrorInvalid(wxT("type"), cfgString, connectionNo);
                  return false;
               }
            }
            else
            {
               printErrorNotFound(wxT("type"), connectionNo);
               return false;
            }
            
            // Look for state.
            if ( wxConfigObject.Read(pattern + wxT("state"), &cfgString) )
            {
               // Check that state is valid.
               if ( !parseState(cfgString, newConnCfg.state) )
               {
                  printErrorInvalid(wxT("state"), cfgString, connectionNo);
                  return false;
               }
            }
            else
            {
               printErrorNotFound(wxT("state"), connectionNo);
               return false;
            }
            
            // Look for device.
            if ( !wxConfigObject.Read(pattern + wxT("device"), &newConnCfg.device) )
            {
               printErrorNotFound(wxT("device"), connectionNo);
               return false;
            }
            
            // If this connection uses serial port, look for port parameters.
            if ( newConnCfg.type == SERIALTNC || newConnCfg.type == SERIALKISSTNC )
            {
               // Look for speed.
               if ( !wxConfigObject.Read(pattern + wxT("speed"), &cfgNumber) )
               {
                  printErrorNotFound(wxT("speed"), connectionNo);
                  return false;
               }
               newConnCfg.portcfg.speed = cfgNumber;
               
               // Look for charlength.
               if ( !wxConfigObject.Read(pattern + wxT("charlength"), &cfgNumber) )
               {
                  printErrorNotFound(wxT("charlength"), connectionNo);
                  return false;
               }
               newConnCfg.portcfg.charlength = cfgNumber;
               
               // Look for parity.
               if ( wxConfigObject.Read(pattern + wxT("parity"), &cfgString) )
               {
                  if ( cfgString.length() != 1 )
                  {
                     printErrorInvalid(wxT("parity"), cfgString, connectionNo);
                     return false;
                  }
                  newConnCfg.portcfg.parity = cfgString.at(0);
               }
               else
               {
                  printErrorNotFound(wxT("parity"), connectionNo);
                  return false;
               }
               
               // Look for stopbit count.
               if ( !wxConfigObject.Read(pattern + wxT("stopbits"), &cfgNumber) )
               {
                  printErrorNotFound(wxT("stopbits"), connectionNo);
                  return false;
               }
               newConnCfg.portcfg.stopbits = static_cast<unsigned int>(cfgNumber);

               // Look for HW flow control.
               if ( !wxConfigObject.Read(pattern + wxT("rtscts"), &cfgString) )
               {
                  printErrorNotFound(wxT("rtscts"), connectionNo);
                  return false;
               }
               else
               {
                  if ( cfgString == wxT("off") )
                  {
                     newConnCfg.portcfg.rtscts = false;
                  }
                  else if ( cfgString == wxT("on") )
                  {
                     newConnCfg.portcfg.rtscts = true;
                  }
                  else
                  {
                     printErrorInvalid(wxT("rtscts"), cfgString, connectionNo);
                     return false;
                  }
               }
               
            }
            
            // Look for error action.
            if ( wxConfigObject.Read(pattern + wxT("erroraction"), &cfgString) )
            {
               if ( cfgString == wxT("log") )
               {
                  newConnCfg.dieOnError = false;
               }
               else if ( cfgString == wxT("terminate") )
               {
                  newConnCfg.dieOnError = true;
               }
               else
               {
                  printErrorInvalid(wxT("erroraction"), cfgString, connectionNo);
                  return false;
               }
            }
            else
            {
               printErrorNotFound(wxT("erroraction"), connectionNo);
               return false;
            }

            // If connection is serial port type and writes to it, look for
            // name, filter, limits and path.
            if ( newConnCfg.type == SERIALKISSTNC && newConnCfg.state != READONLY )
            {
               // Look for name.
               newConnCfg.name = wxT("");
               if ( wxConfigObject.Read(pattern + wxT("name"), &newConnCfg.name) )
               {
                  // Check that name can be used in AX.25 network.
                  char* tmp = fap_check_ax25_call(newConnCfg.name.ToAscii(), 1);
                  if ( !tmp )
                  {
                     printErrorInvalid(wxT("name"), cfgString, connectionNo);
                     return false;
                  }
                  free(tmp);
               }
            
               // Look for filter.
               if ( wxConfigObject.Read(pattern + wxT("txfilter"), &cfgString) )
               {
                  if ( !parseFilter(cfgString, newConnCfg.filterElements) )
                  {
                     printErrorInvalid(wxT("txfilter"), cfgString, connectionNo);
                     return false;
                  }
               }
               else
               {
                  wxLogMessage(wxString::Format(wxT("No txfilter for connection %d, nothing will be gated to it."), connectionNo));
               }
               
               // Look for holdtime.
               newConnCfg.limitercfg.holdtime = 0;
               if ( !wxConfigObject.Read(pattern + wxT("txholdtime"), &newConnCfg.limitercfg.holdtime) )
               {
                  if ( !newConnCfg.filterElements.empty() )
                  {
                     wxLogMessage(wxString::Format(wxT("No holdtime for connection %d, packets gated to this connection will be transmitter without any delay."), connectionNo));
                  }
               }
                        
               // Look for maxrate.
               newConnCfg.limitercfg.amountOfPackets = 0;
               newConnCfg.limitercfg.amountOfTime = 60;
               if ( wxConfigObject.Read(pattern + wxT("txmaxrate"), &cfgString) )
               {
                  string line;
                  line = cfgString.ToAscii();
                  istringstream lineflow (line);
                  string shit;
                  if ( lineflow >> newConnCfg.limitercfg.amountOfPackets &&
                       lineflow >> shit &&
                       lineflow >> newConnCfg.limitercfg.amountOfTime )
                  {
                  }
                  else
                  {
                     printErrorInvalid(wxT("txmaxrate"), cfgString, connectionNo);
                     return false;
                  }
               }
               else
               {
                  if ( !newConnCfg.filterElements.empty() )
                  {
                     printErrorNotFound(wxT("txmaxrate"), connectionNo);
                     return false;
                  }
               }

               // Read rf path.
               if ( wxConfigObject.Read(pattern + wxT("path"), &cfgString) )
               {
                  // Split the path elements and validate them.
                  wxStringTokenizer splitter ( cfgString, wxT(" ") );
                  while ( splitter.HasMoreTokens() )
                  {
                     wxString element = splitter.GetNextToken();
                     char* tmp = fap_check_ax25_call(element.ToAscii(), 1);
                     if ( !tmp )
                     {
                        wxLogMessage( wxT("Error: invalid path element \"") + element + wxT("\".") );
                        return false;
                     }
                     free(tmp);
                     newConnCfg.path.push_back(element);
                  }
               }
            
            } // if ( connection does tx )
            
            // Save the successfully parsed connection.
            configuration.connections.push_back(newConnCfg);
            
            // Prepare for next connection.
            connectionNo++;
            pattern = wxT("/connection");
            pattern << connectionNo;
         }
         
         // As we're here, guess nothing went bad.
         return true;
      }
      
   } //anonymous namespace
   
   ///////////////////////////////////////////////////////////////////////////
   
   // PUBLIC INTERFACE
   
   bool readCfg(const wxString& configFileName, Configuration& configuration)
   {
      // Open config file, quit if opening fails.
      wxFileInputStream cfgFile ( configFileName );
      if ( !(cfgFile.Ok()) )
      {
         // Constructor created nice error message for us, we don't have to.
         return false;
      }
      
      // Turn the config file into a wxConfig object.
      wxFileConfig const wxConfigObject (cfgFile);
      
      // Read callsing.
      if ( !(wxConfigObject.Read(wxT("/general/callsign"), &configuration.callsign)) )
      {
         printErrorNotFound(wxT("callsign"));
         
         return false;
      }
      else
      {
         // Validate callsign.
         char* tmp = fap_check_ax25_call(configuration.callsign.ToAscii(), 1);
         if ( !tmp )
         {
            printErrorInvalid(wxT("callsign"), configuration.callsign );
            return false;
         }
         free(tmp);
      }
      
      // Read latitude
      if ( !(wxConfigObject.Read(wxT("/general/latitude"), &configuration.location.lat)) )
      {
         printErrorNotFound(wxT("latitude"));
         return false;
      }
      else
      {
         if ( abs(configuration.location.lat) > 90  )
         {
            wxLogMessage(wxT("Error: invalid latitude \"") + wxString::Format(wxT("%f"), configuration.location.lat) + wxT("\""));
            return false;
         }
      }
      
      // Read longitude.
      if ( !(wxConfigObject.Read(wxT("/general/longitude"), &configuration.location.lon)) )
      {
         printErrorNotFound(wxT("longitude"));
         return false;
      }
      else
      {
         if ( abs(configuration.location.lon) > 180 )
         {
            wxLogMessage(wxT("Error: invalid longitude \"") + wxString::Format(wxT("%f"), configuration.location.lon) + wxT("\""));
            return false;
         }
      }
      
      // Read possible comment.
      wxConfigObject.Read(wxT("/general/comment"), &configuration.comment);
      if ( configuration.comment.size() > 128 )
      {
         wxLogMessage(wxT("Error: very long comment \"") + configuration.comment +
                      wxT("\"."));
         return false;
      }
      
      // Read beaconing interval.
      if ( !(wxConfigObject.Read(wxT("/general/beaconinterval"), &configuration.beaconInterval)) ||
           configuration.beaconInterval <= 0 )
      {
         wxLogMessage(wxT("Error: beaconing interval invalid or not found."));
         return false;
      }
      // Convert beacon interval to seconds.
      configuration.beaconInterval *= 60;

      // Read duplicate time.
      if ( !(wxConfigObject.Read(wxT("/general/duplicatetime"), &configuration.duplicateTime) ) )
      {
         configuration.duplicateTime = 30;
         //wxLogMessage(wxT("Using default duplicatetime: 30 seconds"));
      }
      else
      {
         if ( configuration.duplicateTime < 0 )
         {
            wxLogMessage(wxT("Error: invalid duplicate time."));
            return false;
         }
      }
      
      // Read lock file directory.
      if ( !(wxConfigObject.Read(wxT("/general/lockdir"), &configuration.lockdir)) )
      {
         configuration.lockdir = wxT("/var/lock");
         //wxLogMessage(wxT("Using default lock file location: /var/lock"));
      }
      else
      {
         // Check for trailing slash.
         if ( configuration.lockdir.Last() == '/' )
         {
            printErrorInvalid(wxT("lockdir"), configuration.lockdir);
            return false;
         }
      }
      
      // Read pid directory.
      if ( !(wxConfigObject.Read(wxT("/general/rundir"), &configuration.rundir)) )
      {
         configuration.rundir = wxT("/var/run");
         //wxLogMessage(wxT("Using default PID file location: /var/run"));
      }
      else
      {
         // Check for trailing slash.
         if ( configuration.rundir.Last() == '/' )
         {
            printErrorInvalid(wxT("rundir"), configuration.rundir);
            return false;
         }
      }
      

      
      // Read uplink settings.
      if ( !readServer(wxConfigObject, configuration) )
      {
         return false;
      }      
      
      // Read connection configurations.
      if ( !readConnections(wxConfigObject, configuration) )
      {
         wxLogMessage(wxT("FAILED"));
         return false;
      }
      // If we're multiport, check that connections have names.
      if ( configuration.connections.size() > 1 )
      {
         for ( unsigned int i = 0; i < configuration.connections.size(); ++i )
         {
            if ( configuration.connections.at(i).name.empty() )
            {
               wxLogMessage(wxString::Format(wxT("Error: connection %d has no name."), i));
               return false;
            }
         }
      }
      
      //Get log file name.
      if ( !(wxConfigObject.Read(wxT("/logging/logfile"), &configuration.logcfg.mainFileName)) )
      {
         wxLogMessage(wxT("Error: generic logfile name invalid or not found."));
         return false;
      }
      
      //Read possible packet log file name.
      wxConfigObject.Read(wxT("/logging/packetlogfile"), &configuration.logcfg.packetFileName);
      
      //Read possible log server iface ip/hostname and port.
      wxString tmp;
      if (wxConfigObject.Read(wxT("/logging/logserver"), &tmp) )
      {
         // Server is needed.
         configuration.logcfg.serverNeeded = true;
         
         // Separate server and port.
         int splitpos = tmp.Find(':');
         if ( splitpos < 0 )
         { 
            wxLogMessage(wxT("Error: log server string \"") + tmp + wxT("\" doesn't contain a ':'."));
            return false;
         }
         else if ( static_cast<unsigned int>(splitpos+1) == tmp.Length() )
         {
            wxLogMessage(wxT("Error: no port number in log server string \"") + tmp + wxT("\"."));
            return false;
         }
         
         // Save hostname.
         wxString host = tmp.Mid(0, splitpos);
         if ( host == wxT("*") )
         {
            configuration.logcfg.packetSrv.AnyAddress();
         }
         else
         {
            configuration.logcfg.packetSrv.Hostname(host);
         }
         
         // Save port.
         wxString port = tmp.Mid(splitpos+1);
         unsigned long int bob;
         if ( !port.ToULong(&bob) )
         {
            wxLogMessage(wxT("Error: invalid port number \"") + port + wxT("\" in log server string \"") + tmp + wxT("\"."));
            return false;
         }
         configuration.logcfg.packetSrv.Service(bob);
      }
      else
      {
         configuration.logcfg.serverNeeded = false;
      }
      
      // File closes when scope ends.
      return true;
   }

   
} //namespace Cfg
