/* $Id: connection.h 321 2009-10-19 14:25:40Z oh2gve $
 *
 * Copyright 2007 Antti Seppnen
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


#ifndef CONNECTION_HH
#define CONNECTION_HH


// own stuff.
#include "cfg.h"
#include "aprspacket.h"

// std stuff.
#include <queue>

// wx stuff.
#include <wx/thread.h>


/// Common interface for any connection via hardware.
/**
 * This class is responsible for all concurrency issues. It uses semaphores to
 * synchronize i/o and mutexes to protect it's critical sections.
 *
 * Functions in the public interface should be called only from the main
 * thread. They have wxASSERTs watching you on that, man.
 */
class Connection : public wxThread
{
 public:
   
   /// Constructor.
   /**
    * \param state State of the connection.
    */
   Connection(const Cfg::ConnectionState& state);

   /// Destructor.
   virtual ~Connection();

   /// Blocks until the connection has been initialized.
   /**
    * Should always be called right after Run(). Use isConnected() to check
    * how the initialization went.
    */
   void waitForConnect();

   /// Get status of the connection.
   /**
    * \return True, if connected. False, if not.
    */
   bool isConnected();

   /// Write given packet to the connection.
   /**
    * Packet is written to the connection as it is.
    * \return True, if writing to output buffer succeeded. False, if not.
    */
   bool write(AprsPacket const* packet);

   /// Read a packet from the connection or timeout.
   /**
    * \param timeout Amount of milliseconds to wait for data.
    * \return A packet on success. NULL on timeout, not connected or some other error happens.
    */
   AprsPacket* readTimeout(unsigned int const& timeout);

   /// Return true, if connection state is not writeonly.
   bool inputEnabled() const;
   
   /// Return true, if connection state is not readonly.
   bool outputEnabled() const;
   
   /// This is the entry point of thread. "Main()"-function.
   /**
    * Must be implemented in child class.
    */
   virtual void* Entry() = 0;

 protected:   //Interface for our child classes.

   /// Change status of this thread to started.
   void threadStarted();
   
   /// Change status this connection to connected.
   /**
    * Raises connected_-flag. Child objects call this before startupDone(), if
    * the startup was successfull.
    */
   void gotConnected();
   
   /// Relase mutex protecting thread's startup.
   /**
    * Child objects call this always after trying to start up.
    */
   void startupDone();
   
   /// Put data received from connection to input queue.
   /**
    * Push data to input buffer and Post() input semaphore. After calling
    * this the main thread can get the packet via readTimeout().
    */
   void inputPush(AprsPacket* data);

   /// Check output queue for data to be transmitted.
   /**
    * Waits until input semaphore tells that data is available or
    * call timeouts.
    * \return NULL if no data was available.
    */
   AprsPacket const* outputPopWait(unsigned int const& timeout);

   /// Change our status to disconnected.
   /**
    * Drops flag connected_ to inform main thread that we're no more connected.
    */
   void gotDisconnected();

 private:

   /// State of this connection.
   const Cfg::ConnectionState state_;
   
   /// Mutex protecting i/o-queues and status flag of this connection.
   /**
    * This is used to serialize operations on these private variables.
    */
   wxMutex mutex_;
   
   /// False as long as we're not running yet.
   bool isStarted_;
   
   /// Status flag of this connection.
   bool connected_;

   /// Semaphore for syncing incoming data flow.
   wxSemaphore* inputSemaphore_;
   /// Semapohre for syncing outgoing data flow.
   wxSemaphore* outputSemaphore_;

   /// Buffer for incoming data.
   std::queue<AprsPacket*>* inputQueue_;
   /// Buffer for outgoing data.
   std::queue<AprsPacket const*>* outputQueue_;
};

#endif // CONNECTION_HH
