/* $Id: logserver.cpp 322 2009-10-19 16:43:34Z oh2gve $
 *
 * Copyright 2009 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Own stuff.
#include "cfg.h"
#include "logserver.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <map>
using std::map;
using std::make_pair;
#include <sstream>
using std::ostringstream;

// wx stuff.
#include <wx/thread.h>
#include <wx/socket.h>
#include <wx/log.h>
#include <wx/string.h>
#include <wx/utils.h>


LogServer::LogServer(wxIPV4address const& cfg, wxString const& ownCall,
                     AprsPacket::Location const& ownLoc):
startupDone_(false), initialized_(true), mutex_(), addr_(cfg),
ownCall_(ownCall), ownLoc_(ownLoc),
masterSocket_(NULL), clients_()
{
   // Create master socket. We want non-blocking i/o.
   masterSocket_ = new wxSocketServer(addr_, wxSOCKET_NOWAIT);
}

LogServer::~LogServer()
{
   // Remove connections.
   for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
   {
      i->second->Destroy();
   }
   
   // Remove master socket.
   if ( masterSocket_ != NULL )
   {
      masterSocket_->Destroy();
   }
}

void* LogServer::Entry()
{
   // Check how master socket creation went. If it wasn't ok, check was it supposed to be done at all.
   if ( masterSocket_ != NULL && masterSocket_->Ok() )
   {
      initialized_ = true;
      wxLogMessage(wxT(" logserver: listening for connections at ") + addr_.Hostname() + wxT(":") + wxString::Format(wxT("%d"), addr_.Service()) );
   }
   else
   {
      initialized_ = false;
      wxLogMessage(wxT(" logserver: Initialization failed."));
   }

   // Now we're running.
   mutex_.Lock();
   startupDone_ = true;
   mutex_.Unlock();
   
   while ( initialized_ && !TestDestroy() )
   {
      // Check socket for incoming connections.
      if ( masterSocket_->WaitForAccept(0, 1) )
      {
         wxLogMessage(wxT(" logserver: Incoming connection detected."));
         wxSocketBase* newClient;
         // Accept the new peer.
         if ( (newClient = masterSocket_->Accept()) != NULL )
         {
            // This requires accessing and eventually altering the client list
            // so we lock it for us.
            mutex_.Lock();
            
            // Get peer data.
            wxIPV4address clientAddr;
            wxString clientAddrStr;
            if ( newClient->GetPeer(clientAddr) )
            {
               clientAddrStr = clientAddr.Hostname();
               clientAddrStr += wxT(":");
               clientAddrStr << clientAddr.Service();
               wxLogMessage(wxT(" logserver: Connetion from ") + clientAddrStr + wxT(" accepted."));
            }
            else
            {
               clientAddrStr = wxT("unknown");
               clientAddrStr << time(0);
               wxLogMessage(wxT(" logserver: Can't get figure out source ip or host of the incoming connection, but accepting it anyway."));
            }
            
            // Save connection.
            clients_.insert(make_pair<wxString, wxSocketBase*>(clientAddrStr, newClient));

            // Greet the log watcher.
            ostringstream greet;
            greet << "# Welcome to the aprsg log server!\r\n";
            greet << "# This is " << string(ownCall_.ToAscii()) << " at " <<
              ownLoc_.lat << " " << ownLoc_.lon << ".\r\n";
            greet << "# Hit q and enter when you want to quit.\r\n";
            newClient->Write(greet.str().c_str(), greet.str().size());
            // Check that correct amount of bytes were written.
            if ( newClient->LastCount() != greet.str().size() )
            {
               wxString num1;
               num1 << newClient->LastCount();
               wxString num2;
               num2 << greet.str().size();
               wxLogMessage(wxT(" logserver: Debug: ") + num1 + wxT(" of ") + num2 + wxT(" bytes could be written to ") + clientAddrStr + wxT("."));
            }
          
            // Client added.
            mutex_.Unlock();
         }
         else
         {
            wxLogMessage(wxT(" logserver: Unidentified error happened while accepting the new connection, we had to discard it."));
         }
      }
      
      // Check if our clients are still connected.
      mutex_.Lock();
      for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
      {
         // Check for disconnected/failed ones. Check input from connected ones.
         if ( i->second->IsDisconnected() || i->second->Error() || i->second->WaitForLost(0,1) )
         {
            if ( i->second->Error() ) logError(i->second->LastError() );
            wxLogMessage(wxT(" logserver: It seems that connection from ") + i->first + wxT(" has been disconnected. Closing it."));
         }
         else
         {
            // Check if user has sent something for us.
            if ( i->second->WaitForRead(0, 1) )
            {
               char byte = 0x00;
               i->second->Read(&byte, 1);
               // Check if we have received a 'q'.
               if ( byte != 'q' )
               {
                  // No 'q' from this connection, we continue with others.
                  continue;
               }
               // We got 'q'.
               string msg = "Have a nice day!\r\n";
               i->second->Write(msg.c_str(), msg.size());
               wxLogMessage(wxT(" logserver: Connection from ") + i->first + wxT(" was closed by the user."));
            }
            else
            {
               // No input from the user.
               continue;
            }
         }
         
         // Okay let's close the connection.
         i->second->Destroy();
         clients_.erase(i);
         // i is now invalid, so don't check any more this time.
         break;
      }
      mutex_.Unlock();
      
   }
   
   // Disconnect client sockets.
   for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
   {
      i->second->Close();
      wxLogMessage(wxT(" logserver: disconnected ") + i->first + wxT("."));
   }
   
   // Wait for death in case of initialization failure.
   while ( !TestDestroy() )
   {
      wxSleep(1);
   }
   
   // Bye.
   return NULL;
}

void LogServer::waitForStartup()
{
   // DEBUG: check that caller of this function is the main thread.
   wxASSERT(IsMain());
   
   bool done = false;
   while ( !done )
   {
      mutex_.Lock();
      done = startupDone_;
      mutex_.Unlock();
   }
}

bool LogServer::operator!()
{
   // DEBUG: check that caller of this function is the main thread.
   wxASSERT(IsMain());
   
   // Check that caller has not skipped calling waitForStartup().
   wxASSERT(startupDone_);
   
   // If we're supposed to do sometring, tell them how initialization went.
   return !initialized_;
}

void LogServer::write(string const& line)
{
   // DEBUG: check that caller of this function is the main thread.
   wxASSERT(IsMain());
   
   // Lock the clients for us.
   mutex_.Lock();
   
   // Send the line to them.
   for ( ClientList::iterator i = clients_.begin(); i != clients_.end(); i++ )
   {
      // Send the line.
      i->second->Write(line.c_str(), line.size());
      // Check that correct amount of bytes were written.
      if ( i->second->LastCount() != line.size() )
      {
         wxString num1;
         num1 << i->second->LastCount();
         wxString num2;
         num2 << line.size();
         wxString num3;
         num3 << i->first;
         wxLogMessage(wxT(" logserver: Debug: ") + num1 + wxT(" of ") + num2 + wxT(" bytes could be written to connection ") + num3 + wxT("."));
      }
      
   }
   
   // We're done.
   mutex_.Unlock();
}

void LogServer::logError(wxSocketError const& lastError) const
{
   wxString error;

   switch (lastError) {
     case wxSOCKET_NOERROR:
      error = wxT("No error happened.");
      break;
     case wxSOCKET_INVOP:
      error = wxT("Invalid operation.");
      break;
     case wxSOCKET_IOERR:
      error = wxT("Input/Output error.");
      break;
     case wxSOCKET_INVADDR:
      error = wxT("Invalid address passed to wxSocket.");
      break;
     case wxSOCKET_INVSOCK:
      error = wxT("Invalid socket (uninitialized).");
      break;
     case wxSOCKET_NOHOST:
      error = wxT("No corresponding host.");
      break;
     case wxSOCKET_INVPORT:
      error = wxT("Invalid port.");
      break;
     case wxSOCKET_WOULDBLOCK:
      error = wxT("The socket is non-blocking and the operation would block.");
      break;
     case wxSOCKET_TIMEDOUT:
      error = wxT("The timeout for this operation expired.");
      break;
     case wxSOCKET_MEMERR:
      error = wxT("Memory exhausted.");
      break;

     default:
      error = wxT("Error 12765. WTF???");
   }

   wxString msg = wxT(" logserver: Debug: ");
   msg += error;
   wxLogMessage(msg);
}
