/* $Id: main-utils.cpp 321 2009-10-19 14:25:40Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 * 
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Own stuff.
#include "main.h"
#include "tnc.h"
#include "kisstnc.h"
#include "ax25iface.h"
#include "uplink.h"
#include "connection.h"
#include "cfg.h"
#include "aprspacket.h"

// std stuff.
#include <vector>
using std::vector;
#include <string>
using std::string;
#include <fstream>
using std::ofstream;
#include <ios>
using std::ios;
#include <sstream>
using std::ostringstream;

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/datetime.h>


// PID FILE HANDLING
bool AprsApp::createPidFile()
{
   pidFile_ = new wxFile;
   wxString name = configuration_.rundir;
   name += wxT("/");
   name += wxT(PACKAGE);
   name += wxT(".pid");
   
   // Check if pidfile already exists.
   if ( pidFile_->Exists(name.GetData()) )
   {
      wxString msg = wxT("Error: another instance of ");
      msg += wxT(PACKAGE);
      msg += wxT(" seems to be running.");
      wxLogMessage(msg);
      
      delete pidFile_;
      pidFile_ = NULL;
      return false;
   }
   
   // Create pid file.
   if ( !pidFile_->Create(name.GetData()) )
   {
      wxLogMessage(wxT("Error: failed to create pid file."));
      delete pidFile_;
      pidFile_ = NULL;
      return false;
   }
   
   // Get our pid.
   wxString pid = wxT("");
   pid << wxGetProcessId();
   pid += wxT("\n");
   
   // Write our pid to pidfile.
   if ( !pidFile_->Write(pid) )
   {
      wxLogMessage(wxT("Error: cannot write to pid file ") + name + wxT("."));
      removePidFile();
      return false;
   }
   
   return true;
}

void AprsApp::removePidFile()
{  
   if ( pidFile_ )
   {  
      pidFile_->Close();
      delete pidFile_;
      pidFile_ = NULL;

      wxString name = configuration_.rundir;
      name += wxT("/");
      name += wxT(PACKAGE);
      name += wxT(".pid");
   
      // We just blindly try to remove the file.
      remove(name.ToAscii());
   }
}


// LOG HANDLING
bool AprsApp::openLog(bool const& initialOpening)
{  
   // Open log file.
   logfilePtr_ = new ofstream(configuration_.logcfg.mainFileName.ToAscii(), ios::app);

   // Check that it opened. Fail if not.
   if ( !(*logfilePtr_) )
   {  
      wxLogMessage(wxT("Error: failed to open log file ") +
                   configuration_.logcfg.mainFileName + wxT("."));
      delete logfilePtr_;
      logfilePtr_ = 0;
      return false;
   }

   // Construct a logging target ant set a nice timestamp format.
   wxLog* wxLogObjectPtr = new wxLogStream(logfilePtr_);
   wxLogObjectPtr->SetTimestamp(wxT("%F %T"));
   
   // Set our log as default logging target. This returns pointer to previous
   // target, which must be deleted.
   delete wxLog::SetActiveTarget(wxLogObjectPtr);
   wxLogMessage(wxT("Log file opened."));

   // If this is first time the log is opened, close stdin, stdout and stderr.
   if ( initialOpening )
   {
      if (dup2(nullFd_, 0) < 0 || dup2(nullFd_, 1) < 0 || dup2(nullFd_, 2) < 0)
      {  
         // Known memory leak here. This has never happened, though.
         wxLogMessage(wxT("Error: Closing stdio failed."));         
         return false;
      }
   }
   
   return true;
}

void AprsApp::closeLog()
{
   if ( logfilePtr_ != NULL )
   {
      // Remove log file from log target and close it.
      wxLogMessage(wxT("Closing log file.\n"));
      delete wxLog::SetActiveTarget(NULL);   
      delete logfilePtr_;
      logfilePtr_ = NULL;
   }
}

bool AprsApp::reopenLog()
{
   if ( logfilePtr_ != NULL )
   {
      closeLog();
      return openLog(false);
   }
   
   return true;
}
      

// CONNECTION HANDLING
bool AprsApp::openConnections()
{  
   // Construct our login line.
   string greet = "user ";
   greet += configuration_.callsign.ToAscii();
   greet += " pass ";
   greet += aprsPass(configuration_.callsign);
   greet += " vers ";
   greet += PACKAGE;
   greet += " ";
   greet += VERSION;
   if ( !configuration_.uplink.filter.IsEmpty() )
   {
      greet += " filter ";
      greet += configuration_.uplink.filter.ToAscii();
   }
   greet += "\r\n";

   // First open the server connection.

   // Create the thread object.
   uplink_ = new Uplink(configuration_.uplink, greet);
   if ( uplink_->Create() != wxTHREAD_NO_ERROR )
   {  
      delete uplink_;
      uplink_ = 0;
      wxLogMessage(wxT("Error: cannot create thread for uplink."));
      return false;
   }
   
   // Start the thread.
   uplink_->Run();
   uplink_->waitForConnect();
   
   // Check that it managed to open it's connection.
   if ( !uplink_->isConnected() )
   {
      wxLogMessage(wxT("Error: uplink could not be opened."));
      
      // Stop the connection thread, and wait for it to die.
      uplink_->Delete();
      uplink_ = 0;
      wxSleep(1);
      return false;
   }
   
   // Uplink successfully opened.
   wxLogMessage(wxT("Uplink opened."));

   
   // Open other connections.

   // Temporary pointer for the connection to be opened.
   Connection* tmpConnPtr = 0;

   // Loop through all given connection configurations.
   for ( unsigned int i = 0; i < configuration_.connections.size(); i++ )
   {  
      // Find out what type of connection we are going to open.
      switch ( configuration_.connections.at(i).type )
      {

       case Cfg::AX25IFACE:
         {
#ifdef HAVE_LIBAX25
            tmpConnPtr = new Ax25Iface(configuration_.connections.at(i).device,
                                            configuration_.connections.at(i).state,
                                            configuration_.callsign);
#else
            wxLogMessage(wxT("Error: AX.25 support is not enabled in this build. Configure --with-ax25 and rebuild to enable it."));
            return false;
#endif
            break;       
         }
       case Cfg::SERIALKISSTNC:
         {
            tmpConnPtr = new KissTnc(configuration_.connections.at(i).device,
                                     configuration_.connections.at(i).state,
                                     configuration_.connections.at(i).portcfg,
                                     configuration_.lockdir);
            break;
         }
       case Cfg::SERIALTNC:
         {
            tmpConnPtr = new Tnc(configuration_.connections.at(i).device,
                                 configuration_.connections.at(i).state,
                                 configuration_.connections.at(i).portcfg,
                                 configuration_.lockdir);
            break;
         }
       default:
         {
            wxString errmsg = wxT("Error: detected unknown connection type for connection ");
            errmsg << i+1;
            errmsg += wxT(" when trying to open it. You have found a bug.");
            wxLogMessage(errmsg);
            return false;
            break;
         }
      }
      
      // Check that port thread got created.
      if ( tmpConnPtr->Create() != wxTHREAD_NO_ERROR )
      {  
         delete tmpConnPtr;
         wxString errmsg = wxT("Error: cannot create thread for connection ");
         errmsg << i+1;
         errmsg += wxT(".");
         wxLogMessage(errmsg);
         return false;
      }
      
      // Start the thread.
      tmpConnPtr->Run();
      tmpConnPtr->waitForConnect();
      
      // If we're interested, check wether thread managed to open connection.
      if ( configuration_.connections.at(i).dieOnError && !tmpConnPtr->isConnected() )
      {
         wxString errmsg = wxT("Error: connection ");
         errmsg << i+1;
         errmsg += wxT(" could not be opened.");
         wxLogMessage(errmsg);
      
         // Stop the connection thread, and wait for it to die.
         tmpConnPtr->Delete();
         wxSleep(1);
         return false;
      }
      
      // Save the port.
      connections_.push_back(tmpConnPtr);

      // Tell that we were succesfull.
      wxString msg = wxT("Connection ");
      msg << i+1;
      msg += wxT(" opened.");
      wxLogMessage(msg);      
   }
   
   return true;
}

void AprsApp::closeConnections()
{
   bool didSumthin = !connections_.empty();
   
   // Close connections in descending order.
   for ( int i = connections_.size()-1; i >= 0; --i )
   {
      // Close the connection and wait for it to go.
      connections_.at(i)->Delete();
      wxSleep(1);
   }
   
   if ( didSumthin )
   {
      wxLogMessage(wxT("Connections closed."));
   }
   
   // Close uplink.
   if ( uplink_ != 0 )
   {
      uplink_->Delete();
      wxSleep(1);
      wxLogMessage(wxT("Uplink closed."));
   }
}

// OTHERS
string AprsApp::aprsPass(wxString callsign)
{
   // Remove possible SSID
   if (callsign.Find('-') != -1)
   {
      callsign = callsign.Mid(0, callsign.Find('-'));
   }
   
   // Convert call to upper case
   callsign = callsign.MakeUpper();
   int result = 0x73e2;
   
   // Go through the whole string and do some XORing.
   for (unsigned int i = 0; i < callsign.Length(); i += 2)
   {
      result ^= (callsign.at(i) << 8);
      result ^= callsign.at(i + 1);
   }
   
   // Convert result from int to string
   ostringstream output;
   output << (result & 0x7fff);
   
   return output.str();
}
