/* $Id: main.cpp 322 2009-10-19 16:43:34Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// own stuff.
#include "main.h"
#include "dupefilter.h"
#include "guard.h"
#include "beacon.h"
#include "packetlog.h"
#include "cfg.h"
#include "aprspacket.h"
#include "packet.h"

// std stuff.
#include <csignal>
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <sstream>
using std::ostringstream;
#include <iomanip>
using std::setw;
#include <map>
using std::map;
using std::make_pair;

// non-std stuff.
#include <fap.h>

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/cmdline.h>


/// A global for OS signal notification.
/**
 * Initialized to false.
 */
int gotSignal = 0;

/// Signal handler for termination signals.
/**
 * Only job is to save the signal into gotSignal.
 * \param sigNumber The received signal.
 */
void handleSignal(int sigNumber)
{
   gotSignal = sigNumber;
}
   

// INITIALIZATION FOR MAIN

bool AprsApp::OnInit()
{
   // Parse cmd line. Possibly does something else.
   if ( !wxAppConsole::OnInit() )
   {
      return false;
   }
      
   // Connect our signal handler to some signals.
   signal( SIGTERM, handleSignal );
   signal( SIGINT, handleSignal );
   signal( SIGUSR1, handleSignal );
   
   // Temporarily initialize libfap for config var validation purposes.
   fap_init();
   
   // Read configuration from file.
   if ( !readCfg(configFileName_, configuration_) )
   {
      wxLogMessage(wxT("There was an error while reading ") + configFileName_ + wxT(", can't start.")); 
      fap_cleanup();
      return false;
   }
   fap_cleanup();
   
   // Initialize null file descriptor for stdio redirection.
   nullFd_ = open("/dev/null", O_RDONLY);
   
   return true;
}


// THE MAIN FUNCTION

int AprsApp::OnRun()
{
   // Fork into daemon process, if requested.
   if ( daemonMode_ )
   {
      int pid = fork();
      if ( pid == 0 )
      {
         // Child process goes here.
      }
      else if ( pid < 0 )
      {
         // Daddy process goes here on failure (no child).
         wxLogMessage(wxT("Error: can't start daemon process."));
         return EXIT_FAILURE;
      }
      else
      {
         // Daddy process goes here on success.
         return EXIT_SUCCESS;
      }
   }
   
   // Create pid file.
   if ( !createPidFile() )
   {
      return EXIT_FAILURE;
   }
   // If in daemon mode, move logging from stdio to a file.
   if ( daemonMode_ && !openLog(true) )
   {
      return EXIT_FAILURE;
   }
   
   /// Initialize APRS parser.
   fap_init();
      
   // Open the devices.
   if ( !openConnections() )
   {
      return EXIT_FAILURE;
   }
   
   // Initialize packet logging.
   Packetlog packetlog(configuration_.logcfg, configuration_.callsign,
                       configuration_.location);
   if ( !packetlog )
   {
      return EXIT_FAILURE;
   }
   
   // Initialize duplicate filtering.
   DupeFilter dupefilter(configuration_.duplicateTime);
   
   // Initialize guards for RF connections.
   vector<Guard*> guards;
   for ( unsigned int i = 0; i < configuration_.connections.size(); ++i )
   {
      guards.push_back( new Guard(connections_.at(i), configuration_.connections.at(i), i+1, configuration_.connections.size()+1,
                                  dupefilter, packetlog, string(configuration_.callsign.ToAscii()) ) );
   }
   
   // Initialize beaconing.
   Beacon beacon(connections_, uplink_, configuration_, dupefilter, packetlog);
   if ( !beacon )
   {
      wxLogMessage(wxString::FromAscii(beacon.error().c_str()));
      for ( unsigned int i = 0; i < guards.size(); ++i )
      {
         delete guards.at(i);
      }
      return EXIT_FAILURE;
   }
   
   
   // Begin gateway action.
   wxLogMessage(wxT("Startup successfull, entering operation mode."));
   bool noError = true;

   while ( noError )
   {
      // Check guards for input.
      for ( unsigned int i = 0; i < guards.size(); ++i )
      {
         // Wait for 10 ms on each guard. We dont' want this to take too long.
         Packet* packet = guards.at(i)->givePacket(10);
         
         // Check that we got something.
         if ( !packet )
         {
            continue;
         }
         
         // Initialize output list.
         packet->outputs = string(configuration_.connections.size()+1, '_');
         
         // Prepare to check the packet. If the packet carried a
         // third-party-packet, we process it instead.
         AprsPacket* data = packet->aprsPacket;
         if ( packet->thirdParty )
         {
            data = packet->thirdParty;
         }
         
         // Check if we can gate this to APRS-IS.
         if ( data->canBeGatedToAprsIs() )
         {
            // Let's create APRS-IS-gateable copy of the packet.
            vector<string> addPath;
            addPath.push_back("qAR");
            addPath.push_back( string(configuration_.callsign.ToAscii()) );
            AprsPacket* tmp = new AprsPacket(*data, addPath);
            
            // Then we send it.
            uplink_->write(tmp);
            packet->outputs.at(0) = 'u';
         }
         else
         {
            // Can't be gated.
            packet->flag = 't';
         }
         
         // Check outputting connections.
         for ( unsigned int o = 0; o < guards.size(); ++o )
         {
            if ( i != o && guards.at(o)->passesFilter(data) )
            {
               // Let's create gated-by-us copy of the packet.
               Packet* out = new Packet(*packet);
               out->inputName = string(configuration_.connections.at(i).name.ToAscii());
               out->aprsPacket = new AprsPacket(*data);
               out->thirdParty = NULL;

               // Hand it to guard.
               guards.at(o)->sendPacket(out);
               
               // Raise log flag for this conn.
               packet->outputs.at(o+1) = 'h';
            }
         }
         
         // Give the packet to log.
         packetlog.write(packet);
      }
      
      // Check uplink for input, waiting a max of 10 ms for it.
      AprsPacket* input = uplink_->readTimeout(10);
      if ( input )
      {
         // Create packet for logging purposes.
         Packet* packet = new Packet();
         packet->input = 'u';
         packet->inputName = "TCPIP";
         packet->aprsPacket = input;
         packet->thirdParty = NULL;
         packet->outputs = string(configuration_.connections.size()+1, '_');
         packet->flag = '-';
         
         // Check for validity.
         string errmsg;
         if ( input->error(errmsg) )
         {
            // It's invalid. Check wether it is a metadata line.
            if ( input->getTnc2Packet().at(0) != '#' )
            {
               packet->flag = 'e';
               packetlog.write(packet);
            }
            else
            {
               delete packet;
               delete input;
            }
         }
         else
         {
            // Check for duplicateness.
            if ( dupefilter.isDupe(input) )
            {
               packet->flag = 'd';
            }
            // Check if the packet can be gated.
            else if ( input->canBeGated() )
            {
               // Check outputting connections.
               for ( unsigned int o = 0; o < guards.size(); ++o )
               {
                  if ( guards.at(o)->passesFilter(input) )
                  {
                     // Let's create gated-by-us copy of the packet.
                     Packet* out = new Packet(*packet);
                     out->aprsPacket = new AprsPacket(*input);
                     out->thirdParty = NULL;
                     
                     // Hand it to guard.
                     guards.at(o)->sendPacket(out);
                     
                     // Raise log flag for this conn.
                     packet->outputs.at(o+1) = 'h';
                  }
               }
            }
            else
            {
               packet->flag = 'n';
            }
            
            // Give the packet to log.
            packetlog.write(packet);
         }
      }
      
      
      // Then let's flush dupe filter.
      dupefilter.flush();
      
      // Keep beaconer awake.
      beacon.kick();
      
      // Make sure that guards are awake.
      for ( unsigned int i = 0; i < guards.size(); ++i )
      {
         guards.at(i)->kick();
      }

      // Check that connections are alive.
      if ( !uplink_->isConnected() )
      {
         wxLogMessage(wxT("Error: uplink has died."));
         noError = false;
         break;
      }
      for ( unsigned int i = 0; i < connections_.size(); ++i )
      {
         if ( configuration_.connections.at(i).dieOnError && !connections_.at(i)->isConnected() )
         {
            wxLogMessage(wxT("Error: connection ") + wxString::Format(wxT("%d"), i) + wxT(" has died."));
            noError = false;
            break;
         }
      }
      
      // Check for signals.
      if ( gotSignal == SIGTERM || gotSignal == SIGTERM )
      {
         // We need to quit.
         break;
      }
      else if ( gotSignal == SIGUSR1 )
      {
         // They want us to reopen files.
         wxLogMessage(wxT("SIGUSR1 received, reopening logs."));
         if ( !packetlog.reopen() || !reopenLog() )
         {
            noError = false;
         }
         else
         {
            wxLogMessage(wxT("Back in business."));
         }
         gotSignal = 0;
      }

   } // Main loop
   
   
   // Put quitting reason in log.
   if ( gotSignal )
   {
      wxLogMessage(wxT("Received SIGINT or SIGTERM from os, quitting."));
   }
   else if ( !noError )
   {
      wxLogMessage(wxT("Fatal error faced, quitting."));
   }
   else
   {
      wxLogMessage(wxT("Shutdown initiated for unknown reason!"));
   }

   // Remove guards.
   for ( unsigned int i = 0; i < guards.size(); ++i )
   {
      delete guards.at(i);
   }
   
   return EXIT_SUCCESS;
}

// Called automagically after OnRun().
int AprsApp::OnExit()
{
   closeConnections();
   closeLog();
   removePidFile();
   
   fap_cleanup();
   
   close(nullFd_);

   // Parent class OnExit() does pretty much nothing?
   return wxAppConsole::OnExit();
}



// INITIALIZATION ASSISTANTS

AprsApp::AprsApp():
daemonMode_(false), configFileName_(), configuration_(), pidFile_(NULL),
logfilePtr_(NULL), nullFd_(0), uplink_(NULL), connections_()
{
}

/// Command line options.
/**
 * Recognized command line options are listed here.
 */
const wxCmdLineEntryDesc CMDLINE[] =
{
   { wxCMD_LINE_SWITCH, wxT("h"), wxT("help"), wxT("Show this help text."), wxCMD_LINE_VAL_STRING, wxCMD_LINE_OPTION_HELP },
   { wxCMD_LINE_SWITCH, wxT("d"), wxT("daemon"), wxT("Run in background, not in terminal."), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
   { wxCMD_LINE_SWITCH, wxT("v"), wxT("version"), wxString(wxT("Prints version of ")) + wxString(wxT(PACKAGE)) + wxString(wxT(" and quits.")), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
   { wxCMD_LINE_PARAM,  wxT("c"), wxT("conf"), wxT("configfile"), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
   //Last line just terminates this list.
   { wxCMD_LINE_NONE }
};

void AprsApp::OnInitCmdLine(wxCmdLineParser& parser)
{
   parser.SetDesc(CMDLINE);
}

bool AprsApp::OnCmdLineParsed(wxCmdLineParser& parser)
{
   // Check if it's version they want.
   if ( parser.Found(wxT("version")) )
   {
      wxLogMessage(wxString(wxT(PACKAGE)) + wxString(wxT(" v")) + wxString(wxT(PACKAGE_VERSION)));
      return false;
   }
   
   // Check if we're asked to go background.
   if ( parser.Found(wxT("daemon")) )
   {
      daemonMode_ = true;
   }
   
   // Look for config file name.
   if( parser.GetParamCount() > 0 )
   {
      configFileName_ = parser.GetParam();
   }
   else
   {
      // Default to this config file name, if it's not found on command line.
      configFileName_ = wxT(SYSCONFDIR);
      configFileName_ += wxT("/");
      configFileName_ += wxT(PACKAGE);
      configFileName_ += wxT(".conf");
   }
   
   return true;
}

// Let the magic happen.
IMPLEMENT_APP_CONSOLE(AprsApp)
