/* $Id: serial.cpp 321 2009-10-19 14:25:40Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "cfg.h"
using Cfg::ConnectionState;
#include "serial.h"

// std stuff.
#include <string>
using std::string;

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>


Serial::Serial(wxString const& deviceName, Cfg::ConnectionState const& state,
                                   Cfg::SerialCfg const& portcfg, wxString const& lockdir):
Connection(state),
lockFileName_(lockdir+wxT("/LCK..")), lockFile_(0),
deviceName_(deviceName), device_(NULL)
{
   // Check that our port isn't already being used by trying to lock it.
   if ( !createLockFile() )
   {
      wxLogMessage(wxT(" serial(") + deviceName + wxT("): failed to create lock file."));
   }
   else
   {
      // Open device.
      device_ = new Serialport(string(deviceName.ToAscii()),
                               portcfg.speed, portcfg.charlength, portcfg.parity, portcfg.stopbits, portcfg.rtscts);
      
      // Check how it dit.
      if ( !(*device_) )
      {
         wxLogMessage(wxT(" serial(") + deviceName + wxT("): failed to open device (") + wxString::FromAscii(device_->error().c_str()) + wxT(")"));
         delete device_;
         device_ = NULL;
      }
      else
      {
         // Successfully opened, change our status.
         gotConnected();
         wxLogMessage(wxT(" serial(") + deviceName + wxT("): device opened."));
      }
   }
}

Serial::~Serial()
{
   // Close the device.
   delete device_;
   
   // Remove lock file of the device.
   removeLockFile();
}


// PROTECTED


bool Serial::readDevice(string& data)
{
   // Check for existance.
   if ( !device_ )
   {
      wxLogMessage(wxT(" serial(") + deviceName_ + wxT("): can't read data, device is not open."));
      return false;
   }
   
   // Check status.
   if ( !(*device_) )
   {
      wxLogMessage( wxT(" serial(") + deviceName_ + wxT("): can't read data, ") + wxString::FromAscii(device_->error().c_str()) );
      delete device_;
      device_ = NULL;
      return false;
   }
   
   // Try to read data.
   if ( !device_->receive(data) )
   {
      wxLogMessage( wxT(" serial(") + deviceName_ + wxT("): reading failed, ") + wxString::FromAscii(device_->error().c_str()) );
      delete device_;
      device_ = NULL;
      return false;
   }
   
   return true;
}

bool Serial::writeToDevice(const std::string& data)
{
   // Check for existance.
   if ( !device_ )
   {
      wxLogMessage(wxT(" serial(") + deviceName_ + wxT("): can't write data, device is not open."));
      return false;
   }
   
   // Check status.
   if ( !(*device_) )
   {
      wxLogMessage( wxT(" serial(") + deviceName_ + wxT("): can't write data, ") + wxString::FromAscii(device_->error().c_str()) );
      delete device_;
      device_ = NULL;
      return false;
   }
   
   // Try to write data.
   if ( !device_->transmit(data) )
   {
      wxLogMessage( wxT(" serial(") + deviceName_ + wxT("): writing failed, ") + wxString::FromAscii(device_->error().c_str()) );
      delete device_;
      device_ = NULL;
      return false;
   }
   
   return true;
}



bool Serial::portIsConnected() const
{
   return device_ != NULL;
}



// PRIVATE

bool Serial::createLockFile()
{
   // First split tty's name from port address.

   // Find last '/'.
   int ind = deviceName_.Find('/', true);
   if ( ind == -1 || ind + 1 >= static_cast<int>(deviceName_.size()) )
   {
      wxString msg = wxT(" serial(") + deviceName_ + wxT("): device name is invalid.");
      wxLogMessage(msg);
      return false;
   }
   
   // Take the rest and add it to lock file name.
   lockFileName_ += deviceName_.Mid(ind+1);
   
   
   // Skip to error handling, if this gets false;
   bool noError = true;
   
   // Check if lock file already exists.
   lockFile_ = new wxFile;
   if ( lockFile_->Exists(lockFileName_.GetData()) )
   {
      wxString msg = wxT(" serial(") + deviceName_ + wxT("): port is already in use.");
      wxLogMessage(msg);
      noError = false;
   }
   

   // Create lock file, if it doesn't exists.
   if ( noError && ! lockFile_->Create(lockFileName_.GetData()) )
   {
      wxString msg = wxT(" serial(") + deviceName_ + wxT("): failed create file \"") + lockFileName_ + wxT("\".");
      wxLogMessage(msg);
      noError = false;
   }
   
   
   // Get our pid.
   wxString pid = wxT("");
   pid << wxGetProcessId();
   pid += wxT("\n");
   
   // Write our pid to lockfile.
   if ( noError && ! lockFile_->Write(pid) )
   {
      wxString msg = wxT(" serial(") + deviceName_ + wxT("): can't write to lock file ") + lockFileName_ + wxT(".");
      wxLogMessage(msg);
      noError = false;
   }
   
   
   // Handle possible error.
   if ( !noError )
   {
      delete lockFile_;
      lockFile_ = 0;
      
      return false;
   }

   
   return true;
}


void Serial::removeLockFile()
{
   if ( lockFile_ != 0 )
   {
      delete lockFile_;
      lockFile_ = 0;

      // We borrow a C feature here. Can a remove happen somehow with std c++ or wxWidgets?
      if ( remove(lockFileName_.ToAscii()) != 0 )
      {
         wxLogMessage(wxT(" serial(") + deviceName_ + wxT("): failed to remove lock file \"") + lockFileName_ + wxT("\"."));
      }
   }
}
