/* $Id: serialport.cpp 319 2009-10-19 14:06:42Z oh2gve $
 * 
 * Copyright 2009 Tapio Aaltonen
 *
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// own stuff.
#include "serialport.h"


// std stuff.
#include <string>
using std::string;
#include <cstring>   // Ye olde string stuff.
#include <cerrno>    // Error number definitions
#include <fcntl.h>   // File control definitions
#include <sstream>
using std::ostringstream;
#include <termios.h> // POSIX terminal control definitions



Serialport::Serialport(string const& devicefile, unsigned int const& baudrate,
                       unsigned int const& charsize, char const& parity, unsigned int const& stopbits,
                       bool const& rtscts):
fd_(-1), error_()
{
   // Open the device.
   fd_ = open(devicefile.c_str(), O_RDWR | O_NOCTTY | O_NDELAY);
   
   // Check for error.
   if ( fd_ == -1 )
   {
      error_ = string("failed to open ") + devicefile + string(" (") + string(strerror(errno)) + string(")");
   }
   else
   {
      // Configure the device blocking.
      fcntl(fd_, F_SETFL, 0);
      
      // Get current config.
      struct termios cfg;
      tcgetattr(fd_, &cfg);
      
      // Enable local mode and receiver.
      cfg.c_cflag |= (CLOCAL | CREAD);
      
      // Select new baudrate.
      switch ( baudrate )
      {
       case 300:
         cfsetispeed(&cfg, B300);
         cfsetospeed(&cfg, B300);
         break;
       case 600:
         cfsetispeed(&cfg, B600);
         cfsetospeed(&cfg, B600);
         break;
       case 1200:
         cfsetispeed(&cfg, B1200);
         cfsetospeed(&cfg, B1200);
         break;         
       case 2400:
         cfsetispeed(&cfg, B2400);
         cfsetospeed(&cfg, B2400);
         break;
       case 4800:
         cfsetispeed(&cfg, B4800);
         cfsetospeed(&cfg, B4800);
         break;
       case 9600:
         cfsetispeed(&cfg, B9600);
         cfsetospeed(&cfg, B9600);
         break;
       case 19200:
         cfsetispeed(&cfg, B19200);
         cfsetospeed(&cfg, B19200);
         break;
       case 38400:
         cfsetispeed(&cfg, B38400);
         cfsetospeed(&cfg, B38400);
         break;
       case 57600:
         cfsetispeed(&cfg, B57600);
         cfsetospeed(&cfg, B57600);
         break;
       default:
         ostringstream errmsg;
         errmsg << "invalid baudrate " << baudrate;
         error_ = errmsg.str();
      }
      
      // Select parity.
      switch ( parity )
      {
       case 'N':
         cfg.c_cflag &= ~PARENB;
         cfg.c_iflag &= ~(INPCK | ISTRIP);
         break;
       case 'E':
         cfg.c_cflag |= PARENB;
         cfg.c_cflag &= ~PARODD;
         cfg.c_iflag |= (INPCK | ISTRIP);
         break;
       case 'O':
         cfg.c_cflag |= PARENB;
         cfg.c_cflag |= PARODD;
         cfg.c_iflag |= (INPCK | ISTRIP);
         break;
       default:
         error_ = string("invalid parity: ") + parity;
      }
      
      // Select amount of databits.
      cfg.c_cflag &= ~CSIZE;
      switch ( charsize )
      {
       case 7:
         cfg.c_cflag |= CS7;
         break;
       case 8:
         cfg.c_cflag |= CS8;
         break;
       default:
         ostringstream errmsg;
         errmsg << "invalid character size " << charsize;
         error_ = errmsg.str();         
      }
      
      // Select stopbit count.
      switch ( stopbits )
      {
       case 1:
         cfg.c_cflag &= ~CSTOPB;
         break;
       case 2:
         cfg.c_cflag |= CSTOPB;
         break;
       default:
         ostringstream errmsg;
         errmsg << "invalid stopbit count: " << stopbits;
         error_ = errmsg.str();
      }
      
      // Select hw flow control status.
      if ( rtscts )
      {
         cfg.c_cflag |= CRTSCTS;
      }
      else
      {
         cfg.c_cflag &= ~CRTSCTS;
      }
      
      // Select raw input.
      cfg.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
      
      // Disable sw flow control.
      cfg.c_iflag &= ~(BRKINT | INLCR | IXON | IXOFF | IXANY);
      
      // Select raw output.
      cfg.c_oflag &= ~OPOST;
      
      // Set read timeout to 100 ms.
      cfg.c_cc[VMIN] = 0;
      cfg.c_cc[VTIME] = 1;
      
      // Set new config.
      tcsetattr(fd_, TCSANOW, &cfg);
   }
   
}



Serialport::~Serialport()
{
   if ( error_.empty() )
   {
      close(fd_);
   }
}



bool Serialport::transmit(string const& data)
{
   // Write given data.
   int n = write(fd_, data.c_str(), data.size());
   
   // Check for errors.
   if ( n < 0 )
   {
      error_ = string("failed to write (") + string(strerror(errno)) + string(")");
      return false;
   }
   else if ( n != static_cast<int>(data.size()) )
   {
      ostringstream errmsg;
      errmsg << "tried to write " << data.size() << " bytes, but wrote " << n;
      error_ = errmsg.str();
      return false;
   }
   
   return true;
}



bool Serialport::receive(string& data)
{
   // Read from device.
   char buffer[512];
   int n = read(fd_, &buffer, 512);
   
   // Check for errors.
   if ( n < 0 )
   {
      if ( errno != EAGAIN )
      {
         error_ = string("failed to read (") + string(strerror(errno)) + string(")");
         return false;
      }
      n = 0;
   }
   
   // Return results.
   data = string(buffer, n);
   return true;
}



bool Serialport::operator!() const
{
   return !error_.empty();
}



string Serialport::error() const
{
   return error_;
}
