/* $Id: uplink.cpp 322 2009-10-19 16:43:34Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen
 * Copyright 2007, 2008, 2009 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Own stuff.
#include "uplink.h"
#include "cfg.h"
using Cfg::UplinkCfg;
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <ctime>

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>



Uplink::Uplink(UplinkCfg const& cfg, string const& login):
Connection(cfg.serverState),
socket_(0), cfg_(cfg), login_(login),
currentServer_(), currentServerNo_(0), lastTrafficTime_(time(0)), lastTryTime_(0),
inputBuffer_()
{
   // Save current server as string.
   currentServer_ = cfg_.serverNames.at(currentServerNo_);
}


Uplink::~Uplink()
{
   if (socket_ != 0)
   {
     socket_->Destroy();
   }
}


void* Uplink::Entry()
{
   // Tell main that we're alive.
   threadStarted();
   
   // Try to connect.
   if (connect())
   {
      //Connection was successfully opened.
      gotConnected();
   }
   startupDone();
   
   // Main loop of this thread.
   while ( !TestDestroy() )
   {
      // Either operate on connection or try to fix it.
      if ( socketIsConnected() )
      {
         // First do rx from the internet.
         
         // Wait for 10 ms to the socket to become readable. This is done
         // regardless of connection state to prevent busy looping.
         if ( socket_->WaitForRead(0, 10) )
         {
            // Read as much data from the socket buffer as possible but not 
            // more than 8 kB.
            char buf[8192];
            socket_->Read(buf, WXSIZEOF(buf));
            
            // If we received bytes, mark this as the latest moment we got data from the socket.
            if ( socket_->LastCount() > 0 )
            {
               lastTrafficTime_ = time(0);
            }
            
            // Save received bytes, if we're supposed to do input.
            if ( inputEnabled() )
            {
               // Append data from buf to storage string character by character.
               for (unsigned int i = 0; i < socket_->LastCount(); i++)
               {
                  inputBuffer_.append(1, buf[i]);
               
                  // When we encounter newline, we have a packet and push it to input buffer.
                  if ( buf[i] == '\n' )
                  {
                     AprsPacket* packet = AprsPacket::fromTnc2Packet(inputBuffer_);
                     inputPush(packet);
                  
                     // Clear the buffer for next packet.
                     inputBuffer_.clear();
                  }
               }
            }
         }
         
         // Then let's do tx to the internet.
         
         // Wait for 1 ms for new data to be written to the socket. Data will be 
         // available at variable "output".
         if ( outputEnabled() )
         {
            AprsPacket const* output = outputPopWait(1);
            if ( output )
            {
               // Get the data to be written.
               string packet = output->getTnc2Packet() + "\r\n";
               // Write it.
               socket_->Write(packet.c_str(), packet.size());
               // Clear the packet.
               delete output;
            }
         }
      }
      else // socketIsConnected() returned false
      {
         // Try to reconnect.
         if ( reconnect() )
         {
            wxLogMessage(wxT(" uplink: connection reopened, normal operation restarted."));
         }
         else
         {
            // If we're not asked to shut down, tell user why we're doing it ayway.
            if ( !TestDestroy() )
            {
               wxLogMessage(wxT(" uplink: unable to recover from error, giving up."));
            }
            break;
         }
      } // if (socketIsConnected())
   } // while (!TestDestroy())
   
   
   // When we get here the socket is no longer connected. Destroy it.
   socket_->Destroy();
   socket_ = 0;

   // Change our status to disconnected.
   gotDisconnected();

   // Wait for main thread to stop this thread.
   while (true)
   {
      if ( TestDestroy() )
      {
         break;
      }
      
      wxMilliSleep(1);
   }
   
   return NULL;
}


bool Uplink::connect()
{
   // Tell user what we're going to do.
   wxLogMessage(wxT(" uplink: trying to connect to ") + currentServer_ + wxT("..."));
   
   // If old socket exists, clear it away.
   if (socket_ != 0)
   {
      socket_->Destroy();
   }
   
   // Create a new socket client.
   socket_ = new wxSocketClient();

   // Connect to address and wait 5 seconds for connection
   socket_->Connect(cfg_.servers.at(currentServerNo_), false);
   socket_->WaitOnConnect(5);
   
   // Check for failures in connection.
   if ( !socket_->IsConnected() || socket_->Error() )
   {
      wxLogMessage(wxT(" uplink: failed."));
      return false;
   }
   
   // Do login.
   socket_->Write(login_.c_str(), login_.size());
   
   // Tell user that we succeeded.
   wxLogMessage(wxT(" uplink: connected successfully! "));
   
   return true;
}


bool Uplink::socketIsConnected()
{
   // Test if the socket exists.
   if (socket_ == 0)
   {
      wxLogMessage(wxT(" uplink: socket has disappeared. (wtf?)"));
      return false;
   }

   // Test if the socket is connected.
   if (!socket_->IsConnected())
   {
      logError();
      return false;
   }

   // Check for I/O errors.
   if (socket_->Error())
   {
      logError();
      return false;
   }
   
   // If enabled, check that we have received data recently.
   if ( cfg_.noTrafficTimeout > 0 )
   {
      if ( time(0) - lastTrafficTime_ > cfg_.noTrafficTimeout )
      {
         wxString msg = wxT(" uplink: No traffic received from ");
         msg += currentServer_;
         msg += wxT(" during last ");
         msg << cfg_.noTrafficTimeout/60;
         msg += wxT(" minutes, seems that connection has died.");
         
         wxLogMessage(msg);
         return false;
      }
   }
         
   return true;
}


bool Uplink::reconnect()
{
   // If we're having an i/o-error or some crazy blocking error, it might be that the connection
   // has just timed out. So let's try to reconnect.

   if ( socket_->LastError() != wxSOCKET_IOERR &&
        socket_->LastError() != wxSOCKET_WOULDBLOCK )
   {
      return false;
   }
      
   // Keep trying until connected or shutdown request comes.
   while ( !TestDestroy() )
   {
      // If we have tried to connect recently, wait a little before trying again.
      if ( lastTryTime_ + cfg_.retryInterval >= time(0) )
      {
         wxString msg = wxT(" uplink: waiting ");
         msg << (lastTryTime_ + cfg_.retryInterval - time(0));
         msg += wxT(" seconds before trying again.");
         wxLogMessage(msg);
      }
      while ( lastTryTime_ + cfg_.retryInterval >= time(0) )
      {
         if ( TestDestroy() )
         {
            return false;
         }
         wxSleep(1);
      }
      
      // Try to connect to all servers.
      for ( unsigned int i = 0; i < cfg_.servers.size(); i++ )
      {
         // Save server number.
         currentServerNo_ = i;
         
         // Save the chosen server into the string.
         currentServer_ = cfg_.serverNames.at(currentServerNo_);
         
         // Flush outgoing packets to avoid delayed gating.
         const AprsPacket* output = outputPopWait(1);
         while ( output != 0 )
         {
            delete output;
            output = outputPopWait(1);
         }
            
         // Try to connect.
         if ( connect() )
         {
            // We're now successfully connected, which means that we can to try to
            // reconnect sooner than after retryInterval (if required).
            lastTryTime_ = 0;
            
            // Last info from the server has just been received, so we can reset that
            // timer too.
            lastTrafficTime_ = time(0);
         
            return true;
         }
      }
      
      // No luck this time.
      lastTryTime_ = time(0);
      
      // Check that our error has not changed to an unhandleable one.
      if ( socket_->LastError() != wxSOCKET_IOERR &&
           socket_->LastError() != wxSOCKET_WOULDBLOCK )
      {
         return false;
      }
   }
   
   return false;
}
   
   
void Uplink::logError() const
{
   wxSocketError lastError = socket_->LastError();
   wxString error;

   switch (lastError) {
     case wxSOCKET_NOERROR:
      error = wxT("No error happened.");
      break;
     case wxSOCKET_INVOP:
      error = wxT("Invalid operation.");
      break;
     case wxSOCKET_IOERR:
      error = wxT("Input/Output error.");
      break;
     case wxSOCKET_INVADDR:
      error = wxT("Invalid address passed to wxSocket.");
      break;
     case wxSOCKET_INVSOCK:
      error = wxT("Invalid socket (uninitialized).");
      break;
     case wxSOCKET_NOHOST:
      error = wxT("No corresponding host.");
      break;
     case wxSOCKET_INVPORT:
      error = wxT("Invalid port.");
      break;
     case wxSOCKET_WOULDBLOCK:
      error = wxT("The socket is non-blocking and the operation would block.");
      break;
     case wxSOCKET_TIMEDOUT:
      error = wxT("The timeout for this operation expired.");
      break;
     case wxSOCKET_MEMERR:
      error = wxT("Memory exhausted.");
      break;

     default:
      error = wxT("Error 12765. WTF???");
   }

   wxString msg = wxT(" uplink: connection to ");
   msg += currentServer_;
   msg += wxT(" failed: ");
   msg += error;
   
   wxLogMessage(msg);
}
